from io import BytesIO, StringIO
import re

from PyPDF2 import PdfFileReader

from arkimedes import convert_date_string_to_iso
from arkimedes.ezid import build_anvl, generate_anvl_strings
from arkimedes.qa import check_lc_naf


def _generate_anvl_from_conservation_report(pdf_data):
    pdf, pdf_url = pdf_data
    pdf_content = StringIO()

    reader = PdfFileReader(BytesIO(pdf))

    for page in reader.pages:
        pdf_content.write(page.extractText())

    text = pdf_content.getvalue().replace("\n", "")

    creator_start_index = text.index("Conservator")
    creator_end_index = text.index("Call Number")
    
    creator = check_lc_naf(
        text[creator_start_index:creator_end_index].split(":")[1].strip()
    )

    title_start_p = re.compile(r"Title:")
    title_start_index = title_start_p.search(text).end()
    title_end_p = re.compile(r"\w+/?\w+:")
    title_end_index = title_end_p.search(text[title_start_index:]).start() + title_start_index

    title = text[title_start_index:title_end_index].strip()

    date_start_p = re.compile("Date of report:")
    date_start_index = date_start_p.search(text).end()
    date_end_p = re.compile("Conservator")
    date_end_index = date_end_p.search(text[date_start_index:]).start() + date_start_index
    date = text[date_start_index:date_end_index].strip()
    
    if date != "":
        date = convert_date_string_to_iso(date)

    return build_anvl(creator, title, date, pdf_url, type_="Text")


def generate_anvl_from_conservation_reports(pdf_data, output_file=None):
    return generate_anvl_strings(
        pdf_data, _generate_anvl_from_conservation_report, output_file
    )



from urllib.parse import quote_plus

from fuzzywuzzy import fuzz
from lxml import etree, html
import requests


def check_lc_naf(name):
    return_name = name
    request_string = f"http://id.loc.gov/search/?q={quote_plus(name)}&q=cs%3ahttp%3a%2f%2fid.loc.gov%2fauthorities%2fnames"
    request = requests.get(request_string)

    if request.ok:
        # collect results from page (let's assume that results on additional
        # pages can't possibly be useful)
        results_xpath = "//tbody[@class='tbody-group']"
        name_xpath = "tr/td/a[@title='Click to view record']"
        matches = []

        min_threshold = 60
        threshold = 70

        results_page = html.fromstring(request.text)

        results = results_page.xpath(results_xpath)

        # see if there are any results
        if results:
            for result in results:
                result_name = result.xpath(name_xpath)[0].text
                ratio = fuzz.ratio(name, result_name)
                if ratio > min_threshold:
                    matches.append((ratio, name, result_name))

        if len(matches) > 1:
            matches.sort(reverse=True)
            one, two = matches[0], matches[1]
            if one[0] == two[0]:
                print(
                    f"Uncertain match. Original: {name}. Top 2 matches: {one} | {two}.\n"
                )
            elif one[0] < threshold:
                print(f"Uncertain match: Original: {name}.  Best match: {one}\n")
            else:
                return_name = matches[0][2]
    else:
        print(f"Could not retrieve {request_string}.")

    return return_name



from inspect import cleandoc
import sys
from time import sleep

import requests


def anvl_to_dict(anvl):
    """Convert single-record ANVL string to dictionary.

    Parameters:
    -----------
    anvl : str
        An ANVL-formatted string.

    Returns:
    --------
    dict
    """
    ark_dict = {}
    for line in anvl.split("\n"):
        if line.startswith("::"):
            ark_dict["ark"] = line[3:]
        else:
            k, v = line.split(":", 1)
            ark_dict[k.strip()] = v.strip()

    return ark_dict


def batch_download(username, password, format_="anvl", compression="zip", *args):
    """Batch download ARKs from EZID.

    Parameters:
    -----------
    username : str
        EZID username.
    password : str
        EZID password.
    format_ : str
        Valid inputs are 'anvl', 'csv', or 'xml'. Defaults to 'anvl'.
    compression : str
        Valid inputs are 'gzip' or 'zip'. Defaults to 'zip'.
    *args : str
        String(s) to pass to EZID API. Strings should follow the 
        format 'key=value'. A full list of parameters can be found here:
        https://ezid.cdlib.org/doc/apidoc.html#parameters

    Returns
    -------
    None
    """
    url = "https://ezid.cdlib.org/download_request"
    headers = {"Content-type": "application/x-www-form-urlencoded"}
    query = f"format={format_}&compression={compression}&{'&'.join(args)}"

    r = requests.post(url, data=query, auth=(username, password), headers=headers)

    if not r.ok or not r.text.startswith("success: "):
        print(f"Request failure!\n----------------\n\n{r.text}")
        sys.exit(1)

    print("Downloading ARK records from EZID..", end="")

    url = r.text[9:]
    file_name = url[url.rfind("/") + 1 :]
    r = requests.get(url)
    while True:
        print(".", end="", flush=True)
        r = requests.get(url)
        if r.status_code == 200:
            with open(file_name, "wb") as fh:
                fh.write(r.content)
            break
        else:
            sleep(5)

    if r.status_code == 200:
        print(file_name)
    else:
        print(f"Download failed\nurl: {url}")


def build_anvl(
    creator,
    title,
    dates,
    target,
    publisher="Iowa State University Library",
    type_="Collection",
    profile="dc",
):
    return cleandoc(
        f"""erc.who:{creator}
            erc.what:{title}
            erc.when:{dates}
            dc.creator:{creator}
            dc.title:{title}
            dc.publisher:{publisher}
            dc.date:{dates}
            dc.type:{type_}
            _target:{target}
            _profile:{profile}
            """
    )


def convert_anvl_file_to_tsv(anvl_file, tsv_file):
    with open(tsv_file, "w", encoding="utf-8") as fh:
        anvls = load_anvl_as_dict(anvl_file)
        #first = next(anvls)
        #header = "\t".join(first.keys())
        #first_row = "\t".join(first.values())
        #fh.write(f"{header}\n{first_row}\n")
        for anvl in anvls:
            fh.write("\t".join(anvl.values()))
            fh.write("\n")


def generate_anvl_strings(data_source, parser, output_file):
    anvls = []

    for d in data_source:
        anvls.append(parser(d))

    if output_file is not None:
        with open(output_file, "w", encoding="utf-8") as fh:
            for a in anvls:
                fh.write(f"{a}\n")
    
    return anvls


def get_value_from_anvl_string(field, anvl):
    anvl_dict = anvl_to_dict(anvl)

    return anvl_dict.get(field)


def load_anvl_as_dict(anvl_file):
    """Creates a generator that yields ARK records as dictionaries.

    Parameters
    ----------
    anvl_file : str or pathlib.Path
        File to load ARK records from.

    Returns
    -------
    generator
        Yields a dictionary where each key, value pair corresponds to an
        ANVL record field.
    """
    ark_dict = {}
    with open(anvl_file, "r", encoding="utf-8") as fh:
        for line in fh:
            line = line.strip()
            if line == "":
                yield ark_dict
            elif line.startswith("::"):
                ark_dict["ark"] = line[3:]
            else:
                k, v = line.split(":", 1)
                ark_dict[k] = v.strip()


def load_anvl_as_str(anvl_file):
    """Creates a generator that yields ARK records as ANVL strings.

    Parameters
    ----------
    anvl_file : str or pathlib.Path
        File to load ARK records from.

    Returns
    -------
    generator
        Yields an ARK record as a string.
    """
    anvl_str = ""
    with open(anvl_file, "r", encoding="utf-8") as fh:
        for line in fh:
            line = line.strip()
            if line == "":
                yield anvl_str
            else:
                anvl_str = "\n".join([anvl_str, line])


def load_anvl_as_str_from_tsv(tsv_file):
    """Creates a generator that yields ANVL strings from a TSV.

    Requires a tab-separated-values file with the following
    headers: dc.creator, dc.title, dc.date, _target, and dc.type.

    Parameters
    ----------
    tsv_file : str or pathlib.Path
        File to load metadata from.

    Returns
    -------
    generator
        Yields an ANVL record as a string.
    """
    with open(tsv_file, "r", encoding="utf-8") as fh:
        keys = fh.readline().strip().split("\t")
        for line in fh:
            values = line.strip("\r\n").split("\t")
            md = dict(zip(keys, values))
            anvl = build_anvl(
                md["dc.creator"],
                md["dc.title"],
                md["dc.date"],
                md["_target"],
                type_=md["dc.type"]
            )

            yield anvl


def load_anvl_as_dict_from_tsv(tsv_file):
    """Creates a generator that yields ANVL dictionaries from a TSV.

    Requires a tab-separated-values file with the following
    headers: dc.creator, dc.title, dc.date, _target, and dc.type.

    Parameters
    ----------
    tsv_file : str or pathlib.Path
        File to load metadata from.

    Returns
    -------
    generator
        Yields an ANVL record as a dictionary.
    """
    for anvl in load_anvl_as_str_from_tsv(tsv_file):
        yield anvl_to_dict(anvl)


def upload_anvl(
    user_name, password, shoulder, anvl_text, action="mint", output_file=None
):
    base_url = "https://ezid.cdlib.org"
    headers = {"Content-Type": "text/plain; charset=UTF-8"}

    if action == "mint":
        request_url = "/".join([base_url, "shoulder", shoulder])
    elif action == "update":
        request_url = "/".join([base_url, "id", shoulder])

    r = requests.post(
        request_url, headers=headers, data=anvl_text.encode("utf-8"), auth=(user_name, password)
    )
    ark = r.text[9:]

    print(anvl_text)
    print(r.text)

    if output_file is not None:
        with open(output_file, "a", encoding="utf-8") as fh:
            fh.write(f":: {ark}\n")
            fh.write(f"{anvl_text.strip()}\n")

    return ark


def view_anvl(user_name, password, ark, print_=True):
    base_url = "https://ezid.cdlib.org/id"
    request_url = "/".join([base_url, ark])
    r = requests.get(request_url, auth=(user_name, password))

    if print_:
        print(r.text)

    return r.text



from io import BytesIO

from lxml import etree

from arkimedes import CONCAT_STRING
from arkimedes.ezid import build_anvl, generate_anvl_strings
from arkimedes.qa import check_lc_naf


def _generate_anvl_from_ead_xml(xml):
    try:
        tree = etree.parse(BytesIO(xml))
        try:
            creator = check_lc_naf(
                tree.xpath(
                    "//archdesc/did/origination[@label='Creator']/*[self::persname or self::corpname]/text()"
                )[0]
            )
        except IndexError:
            creator = ""

        title = tree.xpath("//archdesc/did/unittitle/text()")[0]

        try:
            dates = tree.xpath("//archdesc/did/unitdate/@normal")[0]
        except:
            dates = ""

        target = tree.xpath("//ead/eadheader/eadid/text()")[0]

        return build_anvl(creator, title, dates, target)
    except etree.XMLSyntaxError:
        print(etree.XMLSyntaxError)
        with open("malformed_xml.txt", "a", encoding="utf-8") as fh:
            fh.write(xml.decode("utf-8"))
            fh.write(CONCAT_STRING)


def generate_anvl_from_ead_xml(xml, output_file=None):
    return generate_anvl_strings(xml, _generate_anvl_from_ead_xml, output_file)





"""Classes and functions fork working with the arkimedes database.

To improve various operations, arkimedes tracks ARK records in a local
database. This allows for flexibility in searching through minted ARKs,
tracking local metadata not stored by EZID, and reduces the need to query
the EZID API to get information about existing records.

Classes and functions in this module:

* Ark()
* Base()
* engine()
* Session()

* add_to_db(ark_obj)
* create_db(engine)
* db_exists()
* dump_db(out_file)
* find(filter_col=None, filter=None, session=None)
* find_all(session=None)
* find_ark(ark, session=None)
* find_replaceable(session=None)
* find_url(url, session=None)
* input_is_replaceable(title)
* load_anvl_file_into_db(batch_file, engine=engine, verbose=True)
* sync_db()
* update_db_record(ark, ark_dict)
* url_is_in_db(url)
"""
from copy import copy
from inspect import cleandoc
from pathlib import Path
import re
import sys

from sqlalchemy import Column, create_engine, Integer, String, Text, Boolean
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker

from arkimedes.config import DB_CONN, DB_TYPE, SQLITE_FILE
from arkimedes.ezid import load_anvl_as_dict, load_anvl_as_str

engine = create_engine(DB_CONN)
Session = sessionmaker(bind=engine)
Base = declarative_base()


class Ark(Base):
    """Map ARK records to database rows.

    This class also serves as the schema for the arks table in the database.

    Parameters
    ----------
    ark : str
        Digital object's minted ARK. Primary key for ``arks`` table.
    target : str
        URL the ARK resolves to.
    profile : str
        Preferred metadata profile.
    status : str
        Valid options are: public, reserved, and unavailable.
    owner : str
        Institution owning the ARK.
    ownergroup : str
        Owner group of the ARK.
    created : int
        Unix timestamp of the time the ARK was created.
    updated : int
        Unix timestamp of when the ARK was last changed.
    export : bool
        If the ARK is publicized via external indexing and harvesting
        services.
    dc_creator : str
        Creator of the target resource.
    dc_title : str
        Title of the target resource.
    dc_type : str
        DCMI Type of the target resource.
    dc_date : str
        ISO-8601-formatted date of the target resource's creation.
    dc_publisher : str
        Publishing institution of the target resource.
    erc_when : str
        ISO-8601-formatted date of the target resource's creation.
    erc_what : str
        Title of the target resource.
    erc_who : str
        Creator of the target resource.
    replaceable : bool
        Whether the metadata in a given ARK record can be replaced with
        metadata describing a different resource.

    Attributes
    ----------
    ark : str
        Digital object's minted ARK. Primary key for ``arks`` table.
    target : str
        URL the ARK resolves to.
    profile : str
        Preferred metadata profile.
    status : str
        Valid options are: public, reserved, and unavailable.
    owner : str
        Institution owning the ARK.
    ownergroup : str
        Owner group of the ARK.
    created : int
        Unix timestamp of the time the ARK was created.
    updated : int
        Unix timestamp of when the ARK was last changed.
    export : bool
        If the ARK is publicized via external indexing and harvesting
        services.
    dc_creator : str
        Creator of the target resource.
    dc_title : str
        Title of the target resource.
    dc_type : str
        DCMI Type of the target resource.
    dc_date : str
        ISO-8601-formatted date of the target resource's creation.
    dc_publisher : str
        Publishing institution of the target resource.
    erc_when : str
        ISO-8601-formatted date of the target resource's creation.
    erc_what : str
        Title of the target resource.
    erc_who : str
        Creator of the target resource.
    replaceable : bool
        Whether the metadata in a given ARK record can be replaced with
        metadata describing a different resource.
    """

    __tablename__ = "arks"

    ark = Column(String(30), primary_key=True)
    target = Column(Text)
    profile = Column(String(8))
    status = Column(String(11))
    owner = Column(String(11))
    ownergroup = Column(String(11))
    created = Column(Integer)
    updated = Column(Integer)
    export = Column(Boolean)
    dc_creator = Column(Text)
    dc_title = Column(Text)
    dc_type = Column(String(20))
    dc_date = Column(Text)
    dc_publisher = Column(Text)
    erc_when = Column(Text)
    erc_what = Column(Text)
    erc_who = Column(Text)
    replaceable = Column(Boolean)

    def __repr__(self):
        """How an instance represents itself."""
        return f"<Ark(ark={self.ark})>"

    def from_anvl(self, anvl_string):
        """Populate attributes from a single-record ANVL string.
        
        Parameters
        ----------
        anvl_string : str
            Valid ANVL. See `A Name-Value Language (ANVL) <https://tools.ietf.org/search/draft-kunze-anvl-02>`_
            for more details.

        Returns
        -------
        None
        """
        ark_dict = {
            (p[0].strip() if p[0].strip() != "" else "success"): (
                p[1].strip() if p[0].strip() != "" else p[1][2:].strip()
            )
            for p in [l.split(":", 1) for l in anvl_string.split("\n") if l != ""]
        }

        self.ark = ark_dict.get("success", "")
        self.target = ark_dict.get("_target", "")
        self.profile = ark_dict.get("_profile", "")
        self.status = ark_dict.get("_status", "")
        self.owner = ark_dict.get("_owner", "")
        self.ownergroup = ark_dict.get("_ownergroup", "")
        self.created = int(ark_dict.get("_created", 0))
        self.updated = int(ark_dict.get("_updated", 0))
        self.export = 0 if ark_dict.get("_export") == "no" else 1
        self.dc_creator = ark_dict.get("dc.creator", "")
        self.dc_title = ark_dict.get("dc.title", "")
        self.dc_type = ark_dict.get("dc.type", "")
        self.dc_date = ark_dict.get("dc.date", "")
        self.dc_publisher = ark_dict.get("dc.publisher", "")
        self.erc_when = ark_dict.get("erc.when", "")
        self.erc_what = ark_dict.get("erc.what", "")
        self.erc_who = ark_dict.get("erc.who", "")
        self.replaceable = (
            1
            if ark_dict.get("iastate.replaceable") == "True"
            or input_is_replaceable(ark_dict.get("dc.title", "NOT REPLACEABLE"))
            else 0
        )

    def to_anvl(self):
        """Return attributes as an ANVL string.

        Returns
        -------
        str
        """
        anvl = cleandoc(
            f""":: {self.ark}
                _created: {self.created}
                _export: {"yes" if self.export else "no"}
                _owner: {self.owner}
                _ownergroup: {self.ownergroup}
                _profile: {self.profile}
                _status: {self.status}
                _target: {self.target}
                _updated: {self.updated}
                dc.creator: {self.dc_creator}
                dc.date: {self.dc_date}
                dc.publisher: {self.dc_publisher}
                dc.title: {self.dc_title}
                dc.type: {self.dc_type}
                erc.what: {self.erc_what}
                erc.when: {self.erc_when}
                erc.who: {self.erc_who}
                iastate.replaceable: {self.replaceable}
        """
        )

        return f"{anvl}\n\n"


def add_to_db(ark_obj):
    """Add row to the database's arks table.

    Parameters
    ----------
    ark_obj : arkimedes.db.Ark

    Returns
    -------
    None
    """
    session = Session()
    session.add(ark_obj)
    session.commit()
    session.close()


def create_db(engine):
    """Initialize a new database.
    
    Returns
    -------
    None
    """
    Base.metadata.create_all(engine)


def db_exists():
    """Check if database exists.

    Check if the database arkimedes has been initialized. Currently only
    supports checking for SQLite databases.

    Returns
    -------
    bool
    """
    global DB_TYPE
    global SQLITE_FILE

    if DB_TYPE == "sqlite":
        return Path(SQLITE_FILE).exists()


def dump_db(out_file):
    """Dump database contents to an ANVL file.

    Parameters
    ----------
    out_file: str or pathlib.Path
        File to write database contents to.

    Returns
    -------
    None
    """
    arks = find_all()
    with open(out_file, "w", encoding="utf-8") as fh:
        for ark in arks:
            fh.write(ark.to_anvl())


def find(filter_col=None, filter=None, session=None):
    """Find matching rows in the database.
    
    Finds one or more rows that match the ``filter`` term in the
    ``filter_col``. Currently only supports equality match. filter_col and
    filter require each other. ``session`` is always optional and is most
    useful when an edit function needs to update the results of a query.
    If ``filter_col`` and ``filter`` are not provided, returns the entire
    arks table.

    Parameters
    ----------
    filter_col : str or None
        Table column to search. Optional, but required if using filter.
        Defaults to None.
    filter : str or None
        Search term to use. Optional, but required if using filter_col.
        Defaults to None.
    session : sqlalchemy.orm.Session or None
        Session object to conduct query with. Optional. Defaults to None.

    Returns
    -------
    sqlalchemy.orm.Query
    """
    close_session = False

    if session is None:
        session = Session()
        close_session = True

    if filter is not None:
        results = session.query(Ark).filter(filter_col == filter)
    else:
        results = session.query(Ark)

    if close_session:
        session.close()

    return results


def find_all(session=None):
    """Alias for find() without arguments.

    Parameters
    ----------
    session : sqlalchemy.orm.Session or None
        Session object to conduct query with. Optional. Defaults to None.
    
    Returns
    -------
    sqlalchemy.orm.Query
    """
    return find(session=session)


def find_ark(ark, session=None):
    """Find record by ARK.

    Parameters
    ----------
    session : sqlalchemy.orm.Session or None
        Session object to conduct query with. Optional. Defaults to None.
    
    Returns
    -------
    sqlalchemy.orm.Query
    """
    filter_col = Ark.ark
    return find(filter_col, ark, session)


def find_replaceable(session=None):
    """Find replaceable ARK records.

    Find rows in the arks table where the replaceable column is True.

    Parameters
    ----------
    session : sqlalchemy.orm.Session or None
        Session object to conduct query with. Optional. Defaults to None.
    
    Returns
    -------
    sqlalchemy.orm.Query
    """
    filter_col = Ark.replaceable
    filter = True
    return find(filter_col, filter)


def find_url(url, session=None):
    """Find record by URL in target column.

    Parameters
    ----------
    session : sqlalchemy.orm.Session or None
        Session object to conduct query with. Optional. Defaults to None.
    
    Returns
    -------
    sqlalchemy.orm.Query
    """
    filter_col = Ark.target
    filter = url
    return find(filter_col, filter, session)


def input_is_replaceable(title):
    """Determine of an ARK can be replaced with a new target.

    Occassionally, ARKs have been created for children of a compound object.
    On their own, these ARKs serve no useful purpose. Instead of leaving them,
    we can mark them as replaceable so that we can update already-minted ARKs
    with entirely new metadata and avoid wasting our ARKs.

    Parameters
    ----------
    title : str
        The title of the object in the ARK record. Child objects that don't
        need their own ARKs can be identified by titles like 'Front', 'Back',
        'Page 2', 'p. 7', and even ''.
    
    Returns
    -------
    bool
    """
    bad_title = re.compile(r"(^([Pp](age|\.) \d+|[Ff]ront|[Bb]ack)$|^$)")

    return bool(bad_title.match(title))


def load_anvl_file_into_db(batch_file, engine=engine, verbose=True):
    """Load an ANVL file into database.

    Load an ANVL file that has only fields coresponding the the schema of
    arkimedes.db.Ark into the database. Useful for populating the database
    from ``akimedes.ezid.batch_download`` or restoring from a back-up file.
    Use only for a fresh database as this function does not check to see
    if an record already exists.

    Parameters
    ----------
    batch_file : str or pathlib.Path
        File to load ARK records from.
    engine : sqlalchemy.engine
        Engine to use if the database needs to be created. Defaults to
        ``arkimedes.db.engine.``
    verbose : bool
        If True print the number of the item being added to the database. 

    Returns
    -------
    None
    """
    if not db_exists():
        create_db(engine)

    arks = load_anvl_as_dict(batch_file)
    session = Session()

    for i, a in enumerate(arks):
        if verbose:
            sys.stdout.write(f"\r{i}")
            sys.stdout.flush()

        ark_obj = Ark(
            ark=a.get("ark", ""),
            target=a.get("_target", ""),
            profile=a.get("_profile", ""),
            status=a.get("_status", ""),
            owner=a.get("_owner", ""),
            ownergroup=a.get("_ownergroup", ""),
            created=int(a.get("_created", 0)),
            updated=int(a.get("_updated", 0)),
            export=False if a.get("_export") == "no" else True,
            dc_creator=a.get("dc.creator", ""),
            dc_title=a.get("dc.title", ""),
            dc_type=a.get("dc.type", ""),
            dc_date=a.get("dc.date", ""),
            dc_publisher=a.get("dc.publisher", ""),
            erc_when=a.get("erc.when", ""),
            erc_what=a.get("erc.what", ""),
            erc_who=a.get("erc.who", ""),
            replaceable=True
            if a.get("iastate.replaceable") == "True"
            or input_is_replaceable(a["dc.title"])
            else False,
        )

        session.add(ark_obj)

    session.commit()
    session.close()


def sync_db():
    """Syncronize local database with EZID records.

    Add records from EZID to local database and update local records with
    values from EZID where the two are mismatched.
    """
    pass


def update_db_record(ark, ark_dict):
    """Update existing table row.

    Update the row found by ARK with the values in ark_dict.

    Parameters
    ----------
    ark : str
        ARK of record to edit.
    ark_dict : dict
        Keys map to the fields to update. Values are the new values.

    Returns
    -------
    None

    """
    session = Session()
    record = find_ark(ark, session).first()

    for key, value in ark_dict.items():
        key = key.replace("iastate.", "").replace("_", "").replace(".", "_")
        setattr(record, key, value)

    session.commit()
    session.close()


def url_is_in_db(url):
    """Check if a URL is in the arks table.

    Parameters
    ----------
    url : str
        URL to search for.
    
    Returns
    -------
    bool
    """
    return bool(find_url(url).first())



import requests

CONCAT_STRING = "\n---\n"


def get_from_files(fs, encoding="utf-8"):
    files = []
    for f in fs:
        with open(f, "r", encoding=encoding) as fh:
            content = fh.read()
        files.append(content)

    return files


def get_from_urls(urls):
    files = []
    for url in urls:
        try:
            request = requests.get(url)
            if request.ok:
                files.append(request.content)
        except:
            print(f"Item not found: {url}")

    return files


def get_sources(sources):
    fs = [s for s in sources if not s.startswith("http")]
    urls = [s for s in sources if s.startswith("http")]

    files = get_from_files(fs)
    files.extend(get_from_urls(urls))

    return files


def convert_date_string_to_iso(date_string, date_format="MMDDYYYY", delimiter="/"):
    date_bits = date_string.split(delimiter)

    if date_format == "MMDDYYYY":
        date_out = "-".join(
            [date_bits[2], date_bits[0].zfill(2), date_bits[1].zfill(2)]
        )

    return date_out



"""A command-line script for managing ARKs with the arkimedes library

``arkimedes.py`` provides a command-line script for working with the 
ARKimedes Python library, a library created to manage the creation
and maintenance of ARKs for the Iowa State University Library.

Examples
--------
``arkimedes.py action target [username password] [--source source_list]``

All arkimedes.py commands require an action argument and a target argument.
All commands that interact with the EZID API require an EZID username and
password to be provided.


"""
import argparse

from arkimedes import get_sources
from arkimedes.db import (
    add_to_db,
    Ark,
    Base,
    create_db,
    db_exists,
    engine,
    find_all,
    find_ark,
    find_replaceable,
    find_url,
    Session,
    update_db_record,
    url_is_in_db,
)
from arkimedes.ead import generate_anvl_from_ead_xml
from arkimedes.ezid import (
    anvl_to_dict,
    batch_download,
    get_value_from_anvl_string,
    load_anvl_as_str_from_tsv,
    upload_anvl,
    view_anvl,
)
from arkimedes.pdf import generate_anvl_from_conservation_reports


class MissingArgumentError(Exception):
    pass


def add_ark_to_db(args, ark, anvl):
    ezid_anvl = view_anvl(args.username, args.password, ark, False)
    ark_obj = Ark()
    ark_obj.from_anvl(ezid_anvl)
    add_to_db(ark_obj)


def return_anvl(args, anvl_in):
    if args.source is not None:
        if "\n" in anvl_in:
            anvl = anvl_in
        else:
            with open(anvl_in, "r", encoding="utf-8") as fh:
                anvl = fh.read()
    else:
        raise MissingArgumentError(
            f"Action '{args.action}' requires '--source'."
        )

    return anvl


def submit_md(args, anvl_in):
    anvl = return_anvl(args, anvl_in)

    if args.action == "update":
        ark = upload(args, anvl, "update")
    else:
        url = get_value_from_anvl_string("_target", anvl)

        if url_is_in_db(url):
            print(f"An ARK has already been minted for {url}.\n")
            ark_obj = find_url(url).first()
            view_anvl(args.username, args.password, ark_obj.ark)
        else:
            if args.reuse:
                replaceable = find_replaceable().first()

                if replaceable is not None:
                    args.target = replaceable.ark
                    upload(args, anvl, "update")
                    update_db_record(replaceable.ark, {"iastate.replaceable": False})
                else:
                    upload(args, anvl, "mint")
            else:
                ark = upload(args, anvl, "mint")


def upload(args, anvl, action):
    ark = upload_anvl(
        args.username, args.password, args.target, anvl, action, args.out
    )

    if action == "mint":
        add_ark_to_db(args, ark, anvl)
    else:
        anvl_dict = anvl_to_dict(anvl)
        update_db_record(ark, anvl_dict)
    return ark


def main():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "action",
        help="""Action to take. Accepted arguments are: 'batch-download', 'delete',
'mint-anvl', 'mint-ead', 'mint-conservation-report', 'mint-tsv', 'update', and
 'view'. 'mint-anvl' mints new ARKs from ANVL metadata and 'mint-ead' mints ARKs 
from EAD XML. 'mint-conservation-report' mints new ARKs from Conservation
Report PDFs. 'mint-tsv' mints new ARKs from a TSV file.""",
    )
    parser.add_argument(
        "target",
        help="""After any 'mint-' argument, this must be an ARK shoulder. After
'delete', 'update', or 'view', it must be an ARK. After 'batch' it may be
any character or characters; a hyphen is recommended.""",
    )
    parser.add_argument("username", nargs="?", default="", help="EZID username.")
    parser.add_argument("password", nargs="?", default="", help="EZID password.")
    parser.add_argument(
        "--batch-args",
        nargs="+",
        help="""Additional arguments for batch downloads in 'arg=value' format.
For a full list of available options, see: 
https://ezid.cdlib.org/doc/apidoc.html#parameters""",
    )
    parser.add_argument(
        "--batch-compression",
        help="""File compression to use with batch downloads. Accepted values are
'gzip' and 'zip' If this argument is not given, the default 'zip' is used.""",
    )
    parser.add_argument(
        "--batch-format",
        help="""The file format to be returned when batch downloading ARK records.
Accepted values are 'anvl', 'csv', and 'xml'. If this argument is not given,
the default format 'anvl' is used.""",
    )
    parser.add_argument(
        "--out", help="Output file for recording EZID API response."
    )
    parser.add_argument(
        "--source",
        nargs="+",
        help="Files or URLs that contain metadata for populating ARK records",
    )
    parser.add_argument(
        "--reuse",
        action="store_true",
        help="""When this flag is used with an action that mints a new ARK,
arkimedes will reuse an ARK record that's been marked replaceable, if one
is available, rather than minting a new ARK. If no replaceable record is 
available, a new ARK will be minted.""",
    )

    args = parser.parse_args()

    if not db_exists():
        create_db(engine)

    if args.action == "mint-anvl":
        for anvl in args.source:
            submit_md(args, anvl)
    elif args.action == "mint-ead":
        ead_xml = get_sources(args.source)

        anvls = generate_anvl_from_ead_xml(ead_xml)

        for anvl in anvls:
            submit_md(args, anvl)
    elif args.action == "mint-conservation-report":
        pdfs = get_sources(args.source)
        pdf_data = zip(pdfs, args.source)

        anvls = generate_anvl_from_conservation_reports(pdf_data)

        for anvl in anvls:
            submit_md(args, anvl)
    elif args.action == "mint-tsv":
        for source in args.source:
            for anvl in load_anvl_as_str_from_tsv(source):
                submit_md(args, anvl)

    elif args.action == "batch-download":
        format_ = "anvl"
        compression = "zip"

        if args.batch_format is not None:
            format_ = args.batch_format

        if args.batch_compression is not None:
            compression = args.batch_compression

        if args.batch_args is not None:
            batch_download(
                args.username, args.password, format_, compression, args.batch_args
            )
        else:
            batch_download(args.username, args.password, format_, compression)

    elif args.action == "delete":
        pass
    elif args.action == "update":
        for anvl in args.source:
            submit_md(args, anvl)
    elif args.action == "view":
        view_anvl(args.username, args.password, args.target)


if __name__ == "__main__":
    main()
